package com.imcode.forum.sql;

import com.imcode.db.DataSourceDatabase;
import com.imcode.db.Database;
import com.imcode.db.DatabaseCommand;
import com.imcode.db.DatabaseConnection;
import com.imcode.db.DatabaseException;
import com.imcode.db.commands.InsertIntoTableDatabaseCommand;
import com.imcode.db.commands.SqlQueryDatabaseCommand;
import com.imcode.db.commands.SqlUpdateCommand;
import com.imcode.db.commands.TransactionDatabaseCommand;
import com.imcode.db.commands.UpdateTableWhereColumnEqualsDatabaseCommand;
import com.imcode.db.exceptions.IntegrityConstraintViolationException;
import com.imcode.db.handlers.CollectionHandler;
import com.imcode.db.handlers.SingleObjectHandler;
import com.imcode.forum.Forum;
import com.imcode.forum.ForumPost;
import com.imcode.forum.ForumRepository;
import com.imcode.forum.ForumThread;
import com.imcode.forum.ForumUser;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ddlutils.io.DatabaseIO;
import org.xml.sax.InputSource;

import javax.sql.DataSource;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

public class SqlRepository implements ForumRepository {

    protected Database database;

    public String toString() {
        return ClassUtils.getShortClassName(getClass())+"("+database+")" ;
    }

    public SqlRepository(DataSource dataSource) {
        InputSource forumDdl = new InputSource(getClass().getResourceAsStream("/com/imcode/forum/sql/forum.ddl"));
        org.apache.ddlutils.model.Database databaseDdl = new DatabaseIO().read(forumDdl);
        database = new TableCreatingDatabase(new DataSourceDatabase(dataSource), databaseDdl);
    }

    public Forum getForum(String id) {
        return (Forum) database.execute(
                new SqlQueryDatabaseCommand(
                        "SELECT "+"forum_id" +" from "+"forum_forums" +" WHERE "+"forum_id" +" = ?",
                        new Object[] { id },
                        new SingleObjectHandler(new SqlForumFactory(this))));
    }

    List<ForumThread> getForumThreadsWithNewestPosts(SqlForum forum) {
        return (List<ForumThread>) database.execute(
                new SqlQueryDatabaseCommand(
                        "SELECT ft.thread_id, ft.forum_id\n"
                        + "FROM "+"forum_threads" +" ft\n"
                        + "JOIN (SELECT thread_id, MAX(datetime) datetime FROM forum_posts GROUP BY thread_id) np\n"
                        + "ON ft.thread_id = np.thread_id\n"
                        + "WHERE ft." + "forum_id" + " = ?\n"
                        + "ORDER BY np.datetime DESC\n",
                        new Object[] { forum.getId() },
                        new CollectionHandler(new ArrayList(), new SqlThreadFactory(this, forum)))) ;
    }

    ForumThread createThread(final Forum forum, final ForumPost post) {
        String threadId = (String) database.execute(new TransactionDatabaseCommand() {
            public Object executeInTransaction(DatabaseConnection connection) throws DatabaseException {
                String threadId = ""+new InsertIntoTableDatabaseCommand("forum_threads", new Object[][]{
                        { "forum_id", forum.getId() }
                }).executeOn(connection) ;

                createPost(connection, threadId, null, post);
                return threadId ;
            }

        }) ;
        return new SqlThread(threadId, this, forum.getId(), forum) ;
    }

    public List<Forum> getForums() {
        return (List<Forum>) database.execute(
                new SqlQueryDatabaseCommand(
                        "SELECT forum_id FROM forum_forums",
                        null,
                        new CollectionHandler(new ArrayList(), new SqlForumFactory(this))));
    }

    public Forum createForum(final String forumId) {
        try {
            database.execute(new InsertIntoTableDatabaseCommand("forum_forums", new Object[][] {
                { "forum_id", forumId }
            })) ;
        } catch( DatabaseException dbe ) {
            for ( Throwable t = dbe ; null != t; t = t.getCause() ) {
                if (t instanceof IntegrityConstraintViolationException) {
                    return getForum(forumId) ;
                }
            }
            throw dbe ;
        }
        return new SqlForum(this, forumId);
    }

    List<ForumThread> searchThreads(Forum forum, String[] strings) {
        List whereClauses = new ArrayList() ;
        List parameters = new ArrayList();
        buildSearchClauses(strings, whereClauses, parameters);
        if (null != forum) {
            whereClauses.add("forum_threads.forum_id = ?") ;
            parameters.add(forum.getId()) ;
        }
        String sql = "SELECT forum_threads.thread_id, forum_threads.forum_id" +
                     " FROM forum_posts" +
                     " JOIN forum_threads ON forum_threads.thread_id = forum_posts.thread_id" +
                     (whereClauses.isEmpty() ? "" : " WHERE " + StringUtils.join(whereClauses.iterator(), " AND ")) +
                     " GROUP BY forum_threads.thread_id, forum_threads.forum_id";
        return (List<ForumThread>) database.execute(
                new SqlQueryDatabaseCommand(sql,
                                            parameters.toArray(new String[parameters.size()]),
                                            new CollectionHandler(new ArrayList(), new SqlThreadFactory(this, null)))) ;
    }

    List<ForumPost> searchPosts(Forum forum, String[] strings) {
        List whereClauses = new ArrayList() ;
        List parameters = new ArrayList();
        buildSearchClauses(strings, whereClauses, parameters);
        if (null != forum) {
            whereClauses.add("forum_threads.forum_id = ?") ;
            parameters.add(forum.getId()) ;
        }
        String sql = "SELECT " + SqlPostFactory.POST_COLUMNS +
                     " FROM forum_posts" +
                     " JOIN forum_threads ON forum_threads.thread_id = forum_posts.thread_id" +
                     (whereClauses.isEmpty() ? "" : " WHERE " + StringUtils.join(whereClauses.iterator(), " AND ")) ;
        return (List<ForumPost>) database.execute(
                new SqlQueryDatabaseCommand(sql,
                                            parameters.toArray(new String[parameters.size()]),
                                            new CollectionHandler(new ArrayList(), new SqlPostFactory(this, null)))) ;
    }

    private void buildSearchClauses(String[] strings, List whereClauses, List parameters) {
        if (null == strings) {
            return ;
        }
        for ( String string : strings ) {
            String escapedLikeClause = "%"+string.replaceAll("([!%_\\[])", "!$1")+"%";
            whereClauses.add("( forum_posts.title LIKE ? ESCAPE '!'"
                             + " OR forum_posts.body LIKE ? ESCAPE '!'" 
                             + " OR forum_posts.user_alias LIKE ? ESCAPE '!' )");
            parameters.add(escapedLikeClause);
            parameters.add(escapedLikeClause);
            parameters.add(escapedLikeClause);
        }
    }

    public ForumThread getThread(String threadId) {
        return getSqlThread(threadId, null);
    }

    SqlThread getSqlThread(String threadId, Forum forum) {
        return (SqlThread) database.execute(
                new SqlQueryDatabaseCommand(
                        "SELECT forum_threads.thread_id,forum_threads.forum_id FROM forum_threads WHERE forum_threads.thread_id = ?",
                        new Object[] { threadId },
                        new SingleObjectHandler(new SqlThreadFactory(this, forum))) ) ;
    }

    public ForumPost getPost(String id) {
        return (ForumPost) database.execute(
                new SqlQueryDatabaseCommand("SELECT "+ SqlPostFactory.POST_COLUMNS+" FROM " + "forum_posts" + " WHERE post_id = ?",
                                            new Object[] { id },
                                            new SingleObjectHandler(new SqlPostFactory(this, null)))) ;
    }

    private String createPost(DatabaseConnection connection, String threadId, String parentPostId, ForumPost post) {
        ForumUser user = post.getUser();
        return new InsertIntoTableDatabaseCommand("forum_posts", new Object[][] {
                { "thread_id", threadId },
                { "parent_post_id", parentPostId },
                { "datetime", new Timestamp(post.getDateTime().getTime()) },
                { "user_name", user.getName() },
                { "user_alias", user.getAlias() },
                { "user_email_address", user.getEmailAddress() },
                { "userdata", user.getData() },
                { "title", post.getTitle() },
                { "body", post.getBody() },
                { "hidden", Boolean.valueOf(post.isHidden()) },
        }).executeOn(connection).toString() ;
    }

    void incrementPostViewCount(String postId) {
        database.execute(
                new SqlUpdateCommand(
                        "UPDATE forum_posts SET view_count = view_count + 1 WHERE post_id = ?",
                        new Object[] { postId })) ;
    }

    String createPostReply(final String threadId, final String postId, final ForumPost forumPost) {
       return (String) database.execute(new DatabaseCommand() {
           public Object executeOn(DatabaseConnection connection) throws DatabaseException {
               return createPost(connection, threadId, postId, forumPost );
           }
       }) ;
    }

    SqlPostsSet getThreadPosts(SqlThread thread) {
        return (SqlPostsSet) database.execute(
                new SqlQueryDatabaseCommand("SELECT "+ SqlPostFactory.POST_COLUMNS+" FROM " + "forum_posts" + " WHERE thread_id = ? ORDER BY datetime, post_id",
                                            new Object[] { thread.getId() },
                                            new CollectionHandler(new SqlPostsSet(), new SqlPostFactory(this, thread)))) ;
    }

    void hidePost(String postId, boolean hidden) {
        database.execute(new UpdateTableWhereColumnEqualsDatabaseCommand("forum_posts", "post_id", postId, new Object[][] {
                { "hidden", new Boolean(hidden) }
        })) ;
    }

}
